/*
 *	fft.c Fast Fourie Transfer program
 *	This code is distributed for research purpose only.
 *	Copyright (c) 2016 by NAKATA Hiroshi work@nakata-jp.org .
 *	All rights reserved.
 *
 * Be ware that you must not divide value by 0.
 * It may cause program termination.
 * This progmram may divide value by 0.
 * But GNU C compiler & GNU libraries handles its exception well.
 * The result makes inifinit/-inifinit that is enough for my purpose.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define	TRUE		(-1)	// boolean like value
#define	OUT_SAMPLING_HZ	352800
#define	LENGTH		1024

void	make_sintbl(double *sintbl, int length)
{
	int		i;
	const int	n2 = length/2;
	const int	n4 = length/4;
	const int	n8 = length/8;
	double		c,s,dc,ds,t;

	t = sin(M_PI / length);
	dc = 2 * t * t;
	ds = sqrt(dc * (2 - dc));
	t = 2 * dc;
	c = sintbl[n4] = 1;
	s = sintbl[0] = 0;
	for (i = 1; i < n8; i++) {
		c -= dc;
		dc += t * c;
		s += ds;
		ds -= t * s;
		sintbl[i] = s;
		sintbl[n4 - i] = c;
	}
	if (n8 !=0) {
		sintbl[n8] = sqrt(0.5);
	}
	for (i = 0; i < n4; i++) {
		sintbl[n2 - i] = sintbl[i];
	}
	for (i = 0; i < n2 + n4; i++) {
		sintbl[i + n2] = -sintbl[i];
	}
}

void	make_bitrev(int *bitrev, int length)
{
	const int	half_n = length / 2;
	int		i,j,k;

	for (i = 0, j = 0; TRUE; j += k) {
		bitrev[i++] = j;
		if (i >= length) {
			return;
		}
		k = half_n;
		while (k <= j) {
			j -= k;
			k /= 2;
		}
	}
}

void	fft(double *x, double *y, int length)
{
	int	i,j,h,d,n4;
	int	nn,le1,le,ip;
	double	s,c,dx,dy;

	double	*sintbl = alloca(sizeof(double) * (length + length / 4));
	int	*bitrev = alloca(sizeof(int) * length);

	make_sintbl(sintbl, length);
	make_bitrev(bitrev, length);

	nn = length;
	n4 = length / 4;
	/* -------------- bit reverse ------------------ */
	for (i = 0; i < length; i++) {
		j = bitrev[i];
		if (i < j) {
			double tmp;
			tmp = x[i];
			x[i] = x[j];
			x[j] = tmp;
			tmp = y[i];
			y[i] = y[j];
			y[j] = tmp;
		}
	}
	/* -------------- calculation ------------- */
	le1 = 1;
	for (;;) {
		nn /= 2;
		if (nn < 1) {
			break;
		}
		le = le1 + le1;
		h = 0;
		d = length / le;
		for (j = 0; j < le1; j++) {
			c = sintbl[h + n4];
			s = sintbl[h];
			for (i = j; i < length; i += le) {
				ip = i + le1;
				dx =  x[ip] * c + y[ip] * s;
				dy = -x[ip] * s + y[ip] * c;

				x[ip] = x[i] - dx;
				y[ip] = y[i] - dy;
				x[i] += dx;
				y[i] += dy;
			}
			h += d;
		}
		le1 = le;
	}
}

static inline double	get_dB(double data)
{
	return 10.0 * log(data) / log(10.0);
}

static inline double	get_bits(double data)
{
	return fabs(log(data) / log(2.0) / 2.0);
}

static int	read_index = 0;
float	read_float(FILE *fp)
{
	float	f;

	if (fread(&f, sizeof(float), 1, fp) != 1) {
		fprintf(stderr, "fread() failed index is %d\n", read_index);
		exit(1);
	}
	read_index++;

	return f;
}

int	main(int argc, char *argv[])
{
	int	i;
	int	direct_out;
	float	f;
	float	sum;
	double	real[LENGTH];
	double	image[LENGTH];
	double	energy_all = 0.0;
	double	energy_canhear = 0.0;
	double	energy_noise = 0.0;
	double	energy_noise_canhear = 0.0;
	FILE	*fp;

	if (argc != 2 && argc != 3) {
		fprintf(stderr, "usage:\t%s data_file\n", argv[0]);
		exit(1);
	}

	if (argc == 3) {
		if (strncmp(argv[1], "raw", 3) == 0) ||
		    strncmp(argv[2], "raw", 3) == 0) ) {
			direct_out = 1;
		} else {
			direct_out = 0;
		}
	} else {
		direct_out = 0;
	}

	// read data from file
	fp = fopen(argv[1], "r");
	if (fp == NULL) {
		fprintf(stderr, "file %s open failed\n", argv[1]);
		exit(1);
	}

	// skip top 0.5sec samples (in order to use stable signal)
	for (i = 0; i < OUT_SAMPLING_HZ / 2; i++) {
		f = read_float(fp);
	}
	// skip positive wave
	while (f >= 0.0f) {
		f = read_float(fp);
	}
	// skip negative wave
	while (f < 0.0f) {
		f = read_float(fp);
	}
	//printf("skipped %d data\n", read_index);
	// start at smallest positive value
	real[0] = (double)f;
	image[0] = 0.0;
	sum = f;
	for (i = 1; i < LENGTH; i++) {
		f = read_float(fp);
		real[i] = (double)f;
		image[i] = 0.0;
		sum += f;
	}
	fclose(fp);
	// all data are read.

	if (direct_out) {
		// simply print out input signal (do not FFT)
		printf("sample, strength(energy)\n");
		for (i = 0; i < LENGTH; i++) {
			printf("%d, %le\n", i, real[i]);
		}
		return 0;
	}

	// start FFT
	fft(real, image, LENGTH);
	// get power(single number) from imaginaly value
	for (i = 0; i <= LENGTH / 2; i++) {
		real[i] = real[i] * real[i] + image[i] * image[i];
		energy_all += real[i];		// adds all energy
		if (i != 4) {
			// add energy except original signal
			energy_noise += real[i];
		}
		if (OUT_SAMPLING_HZ * (double)i / LENGTH > 20000) {
			continue;	// skip over 20kHz
		}
		energy_canhear += real[i];	// adds energy under 20kHz
		if (i != 4) {
			// add energy except original signal
			energy_noise_canhear += real[i];
		}
	}
	printf("frequency[Hz], power[dB]\n");
	for (i = 0; i <= LENGTH / 2; i++) {
		printf("%.0lf, %.1lf\n",
			 OUT_SAMPLING_HZ * (double)i / LENGTH,
			 round(get_dB(real[i] / energy_all)));
	}

	energy_noise /= energy_all;
	energy_noise_canhear /= energy_canhear;
	printf("all energy S/N %.2lf [dB] %.2lf [bits]\n",
		 fabs(get_dB(energy_noise)), get_bits(energy_noise));
	printf("under 20kHz energy S/N %.2lf [dB] %.2lf [bits]\n",
		 fabs(get_dB(energy_noise_canhear)),
		 get_bits(energy_noise_canhear));

	return 0;
}
